"""Script to reproduce and demonstrate the bug in the BackendAction class of grid2op."""
from typing import Optional

import grid2op
import numpy as np
from grid2op import Environment
from grid2op.Action import PlayableAction
from grid2op.Converter import IdToAct
from grid2op.Observation import CompleteObservation


class Policy:
    """A simple Policy that simulates the first :num_actions_to_try unitary actions and picks the one with the expected
    lowest max rho of the resulting state.

    :param action_space: The action spaces of the env.
    :param num_actions_to_try: The number of actions to try. (if None -> try all actions)
    """
    def __init__(self, action_space, num_actions_to_try: Optional[int]):
        self.action_space = action_space

        # Initialize the unitary actions candidates.
        self.id_to_act = IdToAct(action_space)
        self.id_to_act.init_converter(set_line_status=False,
                                      change_line_status=False,
                                      set_topo_vect=True,
                                      change_bus_vect=False,
                                      redispatch=False,
                                      curtail=False,
                                      storage=False)
        action_selection_vector = np.asarray(range(self.id_to_act.n), dtype=np.int)

        if num_actions_to_try is not None:
            self.num_actions_to_try = min(num_actions_to_try, len(action_selection_vector))
        else:
            self.num_actions_to_try = len(action_selection_vector)

        self._space_id_to_converter_id = dict()
        for i, converter_id in enumerate(action_selection_vector):
            self._space_id_to_converter_id[i] = converter_id

    def compute_action(self, obs: CompleteObservation, force_bus_action: bool) -> PlayableAction:
        """Compute an action for the given observation by simulating the candidates.

        :param obs: The current observation of the env.
        :param force_bus_action: Specify whether a noop action is allowed.
        :return: A grid2op playable action.
        """

        action_candidates = []
        for action_id in range(self.num_actions_to_try):
            # convert action id to actual action applicable to the grid
            converter_action_id = self._space_id_to_converter_id[action_id]
            action_obj = self.id_to_act.convert_act(converter_action_id)

            # If force_bus_action is set and the action is noop skip this candidate.
            if force_bus_action and action_obj == self.action_space({}):
                continue

            # Simulate this action
            sim_obs, sim_rew, sim_done, sim_info = obs.simulate(action_obj)

            # if simulation is not done, add the action to the candidates.
            if not sim_done:
                action_candidates.append((action_obj, sim_obs.rho.max()))

            # In case the max rho of the simulated observation is below 0.95 stop the enumeration (early stopping)
            if sim_obs.rho.max() < 0.95 and not sim_done:
                break

        # If all candidate actions lead to a done, return noop, otherwise pick the one with the lowest expected max rho.
        if len(action_candidates) == 0:
            top_action = self.action_space({})
        else:
            top_action = list(sorted(action_candidates, key=lambda x: x[1]))[0][0]

        return top_action


def rollout_env(env: Environment, policy: Policy) -> None:
    """Rollout the env, waiting of the bug to occur.

    :param env: The grid2op environment.
    :param policy: A policy to produce actions for a given state.
    """
    done = False
    obs = env.reset()
    step_count = 0
    while not done:
        # Retrieve the lines currently offline in the backend action
        backend_offline_lines = \
            np.where((obs._obs_env._backend_action.current_topo.values[obs.line_or_pos_topo_vect] >= 0) & (
                    obs._obs_env._backend_action.current_topo.values[obs.line_ex_pos_topo_vect] >= 0) == False)[0]

        force_bus_action = False
        # In case lines are offline in the backend but all of them are online in the observation (env), force a bus
        #   splitting action (to make the bug happen)
        if len(backend_offline_lines) > 0 and np.all(obs.line_status[backend_offline_lines]):
            print(f'[{step_count:4}] Offline lines in the backend action: {backend_offline_lines}\n'
                  f'       Line status of these lines {obs.line_status[backend_offline_lines]}')
            force_bus_action = True

        if obs.rho.max() >= 0.98 or force_bus_action:
            action = policy.compute_action(obs, force_bus_action=force_bus_action)
        else:
            action = env.action_space({})

        # Check weather the chosen action will perform a bus splitting action
        execution_performs_bus_action = np.count_nonzero(action.set_bus) > 0

        lines_that_will_go_offline_for_no_reason = list(filter(lambda ll: bool(obs.line_status[ll]),
                                                               backend_offline_lines))

        if execution_performs_bus_action and len(lines_that_will_go_offline_for_no_reason) > 0:
            # Now all lines that have been offline in the backend and are online in the observation will go offline when
            #   a bus splitting action is performed.
            print(f'[{step_count:4}] Lines are offline in the backend, online in the obs -> they will go offline for no'
                  f' reason: {lines_that_will_go_offline_for_no_reason}')

        obs, rew, done, info = env.step(action)
        step_count += 1

        if execution_performs_bus_action and len(lines_that_will_go_offline_for_no_reason) > 0:
            print(f'[{step_count:4}] After step:')
            print(f'\tdone: {done}')
            print(f'\tOffline lines after step: {np.where(~obs.line_status)}')
            print(f'\tDisconnected due to cascading failure {info["disc_lines"]}')
            backend_offline_lines = \
                np.where((obs._obs_env._backend_action.current_topo.values[obs.line_or_pos_topo_vect] >= 0) & (
                        obs._obs_env._backend_action.current_topo.values[obs.line_ex_pos_topo_vect] >= 0) == False)[0]
            print(f'\tbackend lines after: {backend_offline_lines}')
            print(action)
            if not done and np.all(~obs.line_status[lines_that_will_go_offline_for_no_reason]):
                raise Exception(f'Lines ({lines_that_will_go_offline_for_no_reason}) went offline for no reason!')


if __name__ == '__main__':
    """Reproduce the grid2op bug."""

    print(f'Grid2op Version: {grid2op.__version__}')
    environment = grid2op.make('rte_case14_realistic')
    environment.set_id(21)
    environment.seed(1234)
    sim_policy = Policy(environment.action_space, num_actions_to_try=50)
    rollout_env(environment, sim_policy)
